package tools.descartes.prisma.core.scopes;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;

import edu.kit.ipd.descartes.core.NamedElement;
import edu.kit.ipd.descartes.mm.adaptationpoints.AdaptationPointDescriptions;
import edu.kit.ipd.descartes.mm.adaptationpoints.AdaptationpointsFactory;
import tools.descartes.prisma.core.MessageBus;
import tools.descartes.prisma.core.MessageBus.ExchangeType;
import tools.descartes.prisma.core.MessageBus.Listener;
import tools.descartes.prisma.core.ModelRepository;
import tools.descartes.prisma.core.Transaction;
import tools.descartes.prisma.core.Transaction.LockType;
import tools.descartes.prisma.core.adapter.ModelSkeletonAdapter;
import tools.descartes.prisma.core.agent.AgentController;
import tools.descartes.prisma.core.agent.RemoveFilter;
import tools.descartes.prisma.core.exceptions.CommitException;
import tools.descartes.prisma.model.scope.Agent;
import tools.descartes.prisma.model.scope.DelegatedObject;
import tools.descartes.prisma.model.scope.ScopeConfiguration;
import tools.descartes.prisma.model.scope.ScopeFactory;
import tools.descartes.prisma.model.sensor.SensorFactory;
import tools.descartes.prisma.model.sensor.SensorRepository;
import tools.descartes.prisma.model.skeleton.ModelSkeleton;

public abstract class AgentScope extends Scope {

	private static final String SCOPE_CONFIGURATION_SUFFIX = ".scope";
	private static final String ADAPTATIONPOINTS_SUFFIX = ".adaptationpoints";
	private static final String SENSORS_SUFFIX = ".sensors";

	private final String type;
	private ScopeConfiguration configuration;
	private SensorRepository sensors;
	private AdaptationPointDescriptions adaptationPoints;
	private final SystemGlobalScope systemScope;
	private final Map<String, AgentScope> scopeCache = new HashMap<>();
	private final List<NotificationDefinition> notifications = new ArrayList<>();
	private final Map<String, DelegationListener> delegations = new HashMap<>();

	protected class NotificationDefinition {
		private final Class<?> key;
		private final String exchange;
		private final ExchangeType type;

		public NotificationDefinition(Class<?> key, String exchange, ExchangeType type) {
			this.key = key;
			this.exchange = exchange;
			this.type = type;
		}

		public void create() {
			getMessageBus().createExchange(exchange, type);
		}

		public boolean isActive(Object o) {
			return key.isInstance(o);
		}

		public void send(String notificationType, Object o) {
			getMessageBus().publishMessage(exchange, getRoutingKey(o), "",
					new ModelNotification(getModelRepository(), notificationType, key, o));
		}

		public String getRoutingKey(Object o) {
			return ((NamedElement) o).getName();
		}
	}

	protected class DelegationListener implements Listener {
		private AgentController controller;

		public void setDelegationHandler(AgentController controller) {
			this.controller = controller;
		}

		@Override
		public void newMessage(String exchange, String routingKey, Map<String, Object> headers, String message) {
			if (controller != null) {
				ModelNotification notification = new ModelNotification(getModelRepository(), headers);
				EObject instance = notification.getInstance();
				List<String> destinations = controller.getDelegationDestinations(exchange, routingKey, instance);
				for (String curScopeName : destinations) {
					AgentScope curScope = scopeCache.get(curScopeName);
					if (curScope == null) {
						curScope = (AgentScope) getModelRepository().getScope(curScopeName);
						curScope.create();
						scopeCache.put(curScopeName, curScope);
					}
					curScope.registerElement(notification.getInstance());
					curScope.sendNotifications(notification.getNotificationType(), instance);
				}
			}
		}
	}

	public AgentScope(String name, String type, SystemGlobalScope systemScope, ModelRepository repository,
			MessageBus messageBus) {
		super(name, repository, messageBus);
		this.type = type;
		this.systemScope = systemScope;
	}
	
	@Override
	protected String getBasePath() {
		return type;
	}

	public SystemGlobalScope getSystemScope() {
		return systemScope;
	}

	public SensorRepository getSensors() {
		if (sensors == null) {
			sensors = (SensorRepository) getModelRepository().getResource(getSensorsPath()).getContents().get(0);
		}
		return sensors;
	}

	private String getSensorsPath() {
		return type + "/" + getName() + "/" + getName() + SENSORS_SUFFIX;
	}
	
	public AdaptationPointDescriptions getAdaptationPoints() {
		if (adaptationPoints == null) {
			adaptationPoints = (AdaptationPointDescriptions) getModelRepository().getResource(getAdaptationPointsPath()).getContents().get(0);
		}
		return adaptationPoints;
	}
	
	private String getAdaptationPointsPath() {
		return type + "/" + getName() + "/" + getName() + ADAPTATIONPOINTS_SUFFIX;
	}
	
	public ScopeConfiguration getScopeConfiguration() {
		if (configuration == null) {
			configuration = (ScopeConfiguration) getModelRepository().getResource(getScopeConfigurationPath())
					.getContents().get(0);
		}
		return configuration;
	}

	private String getScopeConfigurationPath() {
		return type + "/" + getName() + "/" + getName() + SCOPE_CONFIGURATION_SUFFIX;
	}
	
	@Override
	public boolean create() {
		super.create();
		return super.create();
	}

	@Override
	protected void createResources(Transaction transaction) {
		Resource adaptationPointsResource = getModelRepository().getOrCreateResource(transaction,
				getAdaptationPointsPath());
		if (adaptationPointsResource.getContents().isEmpty()) {
			AdaptationPointDescriptions adaptationPoints = AdaptationpointsFactory.eINSTANCE
					.createAdaptationPointDescriptions();
			adaptationPoints.setName(getName());
			adaptationPointsResource.getContents().add(adaptationPoints);
		}
		
		Resource scopeResource = getModelRepository().getOrCreateResource(transaction, getScopeConfigurationPath());
		if (scopeResource.getContents().isEmpty()) {
			ScopeConfiguration configuration = ScopeFactory.eINSTANCE.createScopeConfiguration();
			scopeResource.getContents().add(configuration);
		}
		
		Resource sensorsResource = getModelRepository().getOrCreateResource(transaction, getSensorsPath());
		if (sensorsResource.getContents().isEmpty()) {
			SensorRepository sensors = SensorFactory.eINSTANCE.createSensorRepository();
			sensorsResource.getContents().add(sensors);
		}
	}

	public Agent getAgent(String id) {
		List<Agent> agents = getScopeConfiguration().getAgents();
		for (Agent curAgent : agents) {
			if (curAgent.getId().equals(id)) {
				return curAgent;
			}
		}
		return null;
	}

	public Agent addAgent(String id) {
		Agent agent = getAgent(id);
		if (agent == null) {
			try (Transaction transaction = getModelRepository().createTransaction()) {
				// Save this agent in the repository
				ScopeConfiguration scopeConfiguration = transaction.attach(getScopeConfiguration());
				transaction.lock(scopeConfiguration, LockType.WRITE);
				agent = ScopeFactory.eINSTANCE.createAgent();
				agent.setId(id);
				scopeConfiguration.getAgents().add(agent);
				transaction.commit();
				
				return transaction.detach(agent);
			} catch (CommitException e) {
				e.printStackTrace();
			}
		}
		return agent;
	}

	public void removeAgent(AgentController agent, RemoveFilter filter) {
		try (Transaction transaction = getModelRepository().createTransaction()) {
			ScopeConfiguration scopeConfiguration = transaction.attach(getScopeConfiguration());
			transaction.lock(scopeConfiguration, LockType.WRITE);

			// IMPORTANT: Clean up all ownerships
			agent.getOwnershipManager().releaseAllOwnerships(transaction, filter);
			scopeConfiguration.getAgents().remove(agent.getAgent());
			transaction.commit();
		} catch (CommitException e) {
			e.printStackTrace();
		}
	}

	public void registerElement(EObject element) {
		// Persist this delegated element in the delegations repository so
		// that agents that will arrive later can still access these elements
		try (Transaction curTransaction = getModelRepository().createTransaction()) {
			ScopeConfiguration scopeConfiguration = curTransaction.attach(getScopeConfiguration());
			curTransaction.lock(scopeConfiguration, LockType.WRITE);

			// Search for existing delegation in the repository.
			for (DelegatedObject curElement : scopeConfiguration.getDelegatedObjects()) {
				if (curElement.getObject().equals(element)) {
					return;
				}
			}

			// Create a new delegation
			DelegatedObject newDelegation = ScopeFactory.eINSTANCE.createDelegatedObject();
			newDelegation.setObject(curTransaction.attach(element));
			scopeConfiguration.getDelegatedObjects().add(newDelegation);
			curTransaction.commit();
		} catch (CommitException e) {
			e.printStackTrace();
		}
	}

	public void deregisterElement(EObject element) {
		// Remove the persisted element delegation from repository
		try (Transaction curTransaction = getModelRepository().createTransaction()) {
			ScopeConfiguration scopeConfiguration = curTransaction.attach(getScopeConfiguration());
			curTransaction.lock(scopeConfiguration, LockType.WRITE);

			// Search for existing delegation in the repository and remove it.
			for (DelegatedObject curElement : scopeConfiguration.getDelegatedObjects()) {
				if (curElement.equals(element)) {
					scopeConfiguration.getDelegatedObjects().remove(curElement);
					break;
				}
			}
			curTransaction.commit();
		} catch (CommitException e) {
			e.printStackTrace();
		}
	}

	public void synchronizeModelSkeleton(AgentController agent, Transaction transaction, ModelSkeleton skeleton) {
		List<ModelSkeletonAdapter> adapters = synchronizeModelSkeletonTree(agent, transaction, skeleton);

		for (ModelSkeletonAdapter curAdapter : adapters) {
			curAdapter.synchronizeContents();
		}
	}

	protected List<ModelSkeletonAdapter> synchronizeModelSkeletonTree(AgentController agent, Transaction transaction,
			ModelSkeleton skeleton) {
		List<ModelSkeletonAdapter> synchronizedAdapters = new LinkedList<>();
		if (skeleton.getAdaptationPoints() != null) {
			ModelSkeletonAdapter adapter = agent.adapt(skeleton.getAdaptationPoints(),
					ModelSkeletonAdapter.class);
			adapter.startTransaction(transaction);
			adapter.synchronizeRoot(getAdaptationPoints());
			synchronizedAdapters.add(adapter);
		}
		if (skeleton.getSensors() != null) {
			ModelSkeletonAdapter adapter = agent.adapt(skeleton.getSensors(),
					ModelSkeletonAdapter.class);
			adapter.startTransaction(transaction);
			adapter.synchronizeRoot(getSensors());
			synchronizedAdapters.add(adapter);
		}
		return synchronizedAdapters;
	}

	public void sendNotifications(String notificationType, EObject object) {
		for (NotificationDefinition curNotification : notifications) {
			if (curNotification.isActive(object)) {
				curNotification.send(notificationType, object);
			}
		}
	}

	public void addNotificationHandler(String exchange, String routingKey, final AgentController agent) {
		String queueName = agent.getAgent().getId() + "." + exchange;
		getMessageBus().createQueue(queueName);
		getMessageBus().bindQueue(queueName, exchange, routingKey);
		getMessageBus().addMessageListener(queueName, new Listener() {
			@Override
			public void newMessage(String exchange, String routingKey, Map<String, Object> headers, String message) {
				agent.handleNotification(new ModelNotification(getModelRepository(), headers));
			}
		});
	}

	public void addDelegationHandler(String exchange, AgentController agent) {
		DelegationListener cur = delegations.get(exchange);
		if (cur != null) {
			cur.setDelegationHandler(agent);
		}
	}

	protected void declareNotification(NotificationDefinition notification) {
		this.notifications.add(notification);
		notification.create();
	}

	protected void declareDelegation(String exchange) {
		// Add a queue for delegation to other scopes.
		String delegationQueueName = exchange + ".Delegation";
		getMessageBus().createQueue(delegationQueueName);
		getMessageBus().bindQueue(delegationQueueName, exchange, "#");
		DelegationListener listener = new DelegationListener();
		getMessageBus().addMessageListener(delegationQueueName, listener);
		delegations.put(exchange, listener);
	}

}
